-- hask2e - translate Haskell types to "English"
-- Copyright (c) 2005 Edwin Steiner <edwin.steiner@gmx.net>
-- 
-- This program is free software; you can redistribute it and/or
-- modify it under the terms of the GNU General Public License
-- as published by the Free Software Foundation; either version 2
-- of the License, or (at your option) any later version.
-- 
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
-- 
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

-- Version: 0.0.2
--      added support for tuples
-- -------- 0.0.1 (first release)

{-
    I wrote this program in an effort to learn the Haskell language.
    Maybe you find it useful | amusing.

    How to use hask2e
    -----------------

    When you start the program it asks you to enter a Haskell type.
    Example:

        edwin@yukon:~/dev/hask2e$ ./hask2e
        Enter a type (nothing to quit):

    Enter a type, as it may occur after '::' in Haskell. (Context
    is currently not supported, so you have to omit the '... =>'
    part, if any.)
    Example:

        [Int] -> a -> Int

    hask2e then tries to translate this into something resembling
    English.
    Example:

        This is a function taking
            a list of Ints
        and returning a function taking
            a value of type `a`
        and returning an Int.

    The IO Monad is treated specially. Try 'IO Int', for example.

    Maybe I'll find the time to add translation of contexts, 
    other standard Monads, etc.

    Have fun,
        -Edwin
-}

module Main
    where

import IO hiding (try)
import Char
import Text.ParserCombinators.Parsec
import qualified Text.ParserCombinators.Parsec.Token as P
import Text.ParserCombinators.Parsec.Language (haskell)

-- lexer for the Haskell language -------------------------------------

lexer  = haskell

whiteSpace = P.whiteSpace lexer
lexeme     = P.lexeme lexer
symbol     = P.symbol lexer
natural    = P.natural lexer
parens     = P.parens lexer
semi       = P.semi lexer
identifier = P.identifier lexer
reserved   = P.reserved lexer
reservedOp = P.reservedOp lexer

-- a simple AST for types ---------------------------------------------

data Type
    = TyFun { tyfunArgBType :: Type, tyfunResType :: Type }
    | TyApp { tyappBType :: Type, tyappAType :: Type }
    | TyCon { tycon :: String }
    | TyVar { tyvar :: String }
    | TyList { tylistElmType :: Type }
    | TyTup { tytupTypes :: [Type] }
    deriving (Show)

-- an element of the generated explanation ----------------------------

data Explanation
    = Text String
    | Break
    | Indent
    | Outdent
    deriving (Show)

-- context passed down to the explanation functions -------------------

data ExplainContext
    = ExCtx { ctxValue :: Bool   -- phrase it as a value (rather than type)
            , ctxPlural :: Bool  -- make it plural
            }

-- Parsec parser for the type syntax ----------------------------------

parseConid = do
                first <- upper
                rest  <- many (upper <|> lower <|> digit <|> char '\'')
                whiteSpace
                return (first:rest)

parseVarid = do
                first <- lower
                rest  <- many (upper <|> lower <|> digit <|> char '\'')
                whiteSpace
                return (first:rest)

parseQtycon = do
                id <- parseConid
                id2 <- option "" (do{ symbol "."
                                    ; tycon <- parseQtycon
                                    ; return ('.':tycon) })
                return (id ++ id2)

parseGtycon = parseQtycon
            <|> do { char '('
                   ; inner <- (string "->" <|> many (char ','))
                   ; char ')'
                   ; whiteSpace 
                   ; return ("(" ++ inner ++ ")") }
            <|> symbol "[]"

parseAType = try (do { conid <- parseGtycon <?> "type constructor"
                     ; return TyCon { tycon = conid }
                     })
            <|> do
                varid <- parseVarid
                return TyVar { tyvar = varid }
            <|> do
                symbol "["
                elem <- parseType
                symbol "]"
                return TyList { tylistElmType = elem }
            <|> do
                symbol "("
                inner <- parseType `sepBy1` (symbol ",")
                symbol ")"
                if length(inner) == 1
                    then return(head inner)
                    else return(TyTup { tytupTypes = inner })

parseBType = do
                atypes <- parseAType `sepBy1` whiteSpace
                return(typeApplication(reverse atypes)) -- XXX this is ugly, but it works
                
parseType = do
                btype <- parseBType
                funres <- option Nothing (do{ symbol("->")
                                            ; r <- parseType
                                            ; return(Just r)
                                            })
                case funres of
                    Just r -> return $ TyFun { tyfunArgBType = btype,
                                               tyfunResType = r }
                    Nothing -> return btype

-- typeApplication: helper function, see parseBType
typeApplication [x] = x
typeApplication (x:xs) = TyApp { tyappAType = x, 
                                 tyappBType = typeApplication xs }

-- make an explanation consisting of a single Text element
txt :: String -> [Explanation]
txt str = [Text str]

-- return True if the String starts with a vocal
startsWithVocal :: String -> Bool
startsWithVocal "" = False
startsWithVocal (x:xs) =
    case Char.toLower(x) of
        'a' -> True
        'e' -> True
        'i' -> True
        'o' -> True
        'u' -> True
        _   -> False

-- prefix String with a/an
withparticle :: String -> String
withparticle str =
    if startsWithVocal(str)
        then "an " ++ str
        else "a " ++ str

-- explanation functions ----------------------------------------------

{- These functions transform the AST into a list of Explanation elements. -}

explain :: Type -> ExplainContext -> [Explanation]

explain TyVar { tyvar=v } ctx = 
    if ctxValue(ctx)
        then if ctxPlural(ctx)
            then txt "values of type " ++ e
            else txt "a value of type " ++ e
        else txt "type " ++ e
    where
        e = txt ("`" ++ v ++ "`")

explain TyCon { tycon=c } ctx = 
    if ctxValue(ctx)
        then if ctxPlural(ctx)
            then txt $ c ++ "s"
            else txt $ withparticle c
        else txt c

explain TyApp { tyappBType=TyCon { tycon="IO" }, tyappAType=a } ctx =
    if ctxValue(ctx)
        then if ctxPlural(ctx)
            then txt "IO actions returning " ++ explain a ctx'
            else txt "an IO action returning " ++ explain a ctx'
        else txt "IO action returning " ++ explain a ctx'
    where
        ctx' = ctx { ctxPlural=False, ctxValue=True }

explain TyApp { tyappBType=b, tyappAType=a } ctx =
            explain b ctx1 ++ txt " with " ++ explain a ctx2
    where
        ctx1 = ctx
        ctx2 = ctx { ctxPlural=False, ctxValue=False }

explain TyFun { tyfunArgBType=a, tyfunResType=r } ctx =
    if ctxValue(ctx)
        then if ctxPlural(ctx)
            then txt "functions taking" ++ arg ++ txt "and returning " ++ ret
            else txt "a function taking" ++ arg ++ txt "and returning " ++ ret
        else txt "function that takes" ++ arg ++ txt "and returns " ++ ret
    where
        arg = [Break,Indent] ++ explain a ctx {ctxPlural=False,ctxValue=True} ++ [Break,Outdent]
        ret = explain r ctx {ctxPlural=False,ctxValue=True}

explain TyList { tylistElmType=e } ctx = 
    if ctxValue(ctx)
        then if ctxPlural(ctx)
            then txt "lists of " ++ elm
            else txt "a list of " ++ elm
        else txt "list of " ++ elm
    where
        elm = explain e ctx { ctxPlural=True,ctxValue=True }

explain TyTup { tytupTypes=ts } ctx =
    if ctxValue(ctx)
        then if ctxPlural(ctx)
            then txt "tuples of " ++ e
            else txt "a tuple of " ++ e
        else txt "tuple of " ++ e
    where
        e = [Break,Indent] ++ concatMapWithLast (\t -> explain t ctx' ++ txt "," ++ [Break]) 
                                                (\t -> [Outdent,Text "and ",Indent] ++ explain t ctx')
                                                ts ++ [Outdent]
        ctx' = ctx { ctxPlural=False, ctxValue=True }

concatMapWithLast :: (a -> [b]) -> (a -> [b]) -> [a] -> [b]
concatMapWithLast f flast [] = []
concatMapWithLast f flast [x] = flast x
concatMapWithLast f flast (x:xs) = f x ++ concatMapWithLast f flast xs

-- functions for printing the Explanation list ------------------------

indentation 0 = ""
indentation indent = "    " ++ indentation (indent-1)

printExplanation :: [Explanation] -> Int -> Int -> IO ()

printExplanation [] indent len = do
    return ()

printExplanation ((Text str):es) indent len = do
    if len == 0
        then putStr(indentation(indent))
        else putStr("")
    putStr(str)
    printExplanation es indent (len + length(str))

printExplanation (Break:es) indent len = do
    if len /= 0
        then do putStrLn("")
                printExplanation es indent 0
        else printExplanation es indent len

printExplanation (Indent:es) indent len = do
    printExplanation es (indent+1) len

printExplanation (Outdent:es) indent len = do
    printExplanation es (indent-1) len

-- main program -------------------------------------------------------

main :: IO ()
main = do
    hSetBuffering stdin LineBuffering
    mainLoop

mainLoop = do
    putStrLn "Enter a type (nothing to quit):"
    line <- getLine
    putStrLn ""
    if line == ""
        then return ()
        else do
            explainType line
            mainLoop

explainType sig = do
    case (parse parseType "" sig) of
        Left err -> print err
                    -- let ... context, then
                    -- for (an) arbitrary type(s) {free variables}
                    -- 'foo' is ...
        Right t  -> printExplanation expl 0 0
            where
                expl = txt "This is " ++ explain t ExCtx { ctxValue=True, ctxPlural=False } 
                                      ++ txt "." ++ [Break]
    putStrLn ""

